{{
littlebigspin.spin - an external memory access, fetch, and run spin interpreter.
Copyright (c) 2011, John Steven Denson (jazzed) - extensive changes for external memory.
Copyright (c) 2004, Parallax, Inc. - original interpreter code.

**************************************************************************************

License terms are currently ALL RIGHTS RESERVED until Parallax approves general usage.

**************************************************************************************
}}

' if not BIGSPIN and not NORMALSPIN, it's BIGSPIN without external memory ... sort of

#ifndef BIGSPIN
#define BIGSPIN
#endif

#ifndef SPINSIM
'#define SPINSIM
#endif

#ifndef BMA
'#define BMA
#endif

#define C3
'#define DRACBLADE
'#define SDRAM

{{
This is a port of the original PNut Interpreter for use with external memory.
At this time only 64KB of external memory is usable - thus "little big spin"

This interpreter uses overlays for spin code and performance could be enhanced
marginally by using precalculations of overlay lengths and more HUB variable sharing.

For the time being, a binary is included in this file just for test purposes
to load the external memory. Nothing would stop a user from loading the image
before running the interpreter.

All variables including stack are kept in external memory to allow any size
spin program to run whenever 32 bit registers are enabled.
}}

'        *****************************************
'        *                                        *
'        *        PNut Interpreter                *
'        *                                        *
'        *        Version 0.1   12/10/2004        *
'        *                                        *
'        *        (C) 2004 Parallax, Inc.         *
'        *                                        *
'        *****************************************
'


con
  _clkmode      = u#_clkmode
  _xinfreq      = u#_xinfreq

  EOL           = u#EOL
  BUFSIZE       = 2048

obj

#ifdef C3
    u : "userdefs_c3"
    f : "fsrwFemto_c3"
    c : "spisram_cache_C3"
#endif
#ifdef DRACBLADE
    u : "userdefs"
    f : "fsrwFemto_c3"
    c : "sram_cache_dracblade"
#endif
#ifdef SDRAM
    u : "userdefs"
    f : "fsrw"
    c : "SdramCache"
#endif
#ifdef SPINSIM
    s : "conio"
    f : "fileio"
    c : "SimCache"
#else
    s : "FullDuplexSingleton"
#endif
    'd : "BMAUtility"
    b : "BigHelper"

dat
'
' BigHack WARNING
'
' padbuffer reserves space for COG communications structures
' This is a hack to allow a demo to work. By adding this, it
' pushes the interpreter's communications infrastructure into
' upper hub memory so that devices have room to communicate
' with COG interfaces without interference.
' A real solution is needed.
'
  padbuffer     byte 0 [$1000]

  cogbuf        long 0 [$1f0]       ' COG buffer
  running       long 0
  buffer        byte 0 [BUFSIZE+1]  ' SDCARD buffer

#ifdef C3
  paddit        long 0 [2] ' current cache driver needs a page aligned address
  cache		byte 0 [8192]
  ioControl     long 0 [2] ' SD parameters
#endif
#ifdef DRACBLADE
  paddit        long 0 [2] ' current cache driver needs a page aligned address
  cache		byte 0 [8192]
  ioControl     long 0 [2] ' SD parameters
#endif
#ifdef SDRAM
  cache		long 0 [c#CACHESIZE]
#endif

dat

  cachecmd      long 0          ' args must be in this order
  cachedata     long 0
  xmpbase       long 0
  xmvbase       long 0
  xmdbase       long 0
  xmpcurr       long 0
  xmdcurr       long 0


pub main
'   start(string("test.bin"))    ' works. calls object method
'   start(string("ctest.bin"))   ' works. calls peer method
'   start(string("cogtog.bin"))  ' works. cognew a simple P28 pin toggler
'   start(string("cogctrl.bin")) ' works. cognew a P28 pin toggler low until ctrl != 0
#ifdef C3
    start(string("hello_c3.bin"))' works.
#endif
#ifdef DRACBLADE
    start(string("hello_drac.bin"))' unknown
#endif
#ifdef SDRAM
    start(string("hello.bin"))   ' works.
#endif

pub start(filename) | n, len, val, msize

  ifnot running

    running++   ' never replace the interpreter

    s.start(31,30,0,115200)
    waitcnt(clkfreq+cnt)
    s.str(string(u#EOL,"Starting ... "))

#ifdef BIGSPIN
    c.startx(@cachecmd,@cache)
#endif

#ifdef C3
    if f.startSD(@ioControl) == 0
        s.str(string(u#EOL,"SDCARD Start Error ... "))
        repeat
#endif
    result := \f.mount_explicit(u#spiDO, u#spiClk, u#spiDI, u#spiCS)

    if result
        s.str(string(u#EOL,"SDCARD Mount Error ... "))
        repeat

    result := \f.popen(filename,"r")
    if result
        s.str(string(u#EOL,"SDCARD File Open for Read Error ... "))
        repeat

#ifdef BIGSPIN

    len    := 0
    result := 0
    repeat
        result := f.pread(@buffer, BUFSIZE)
        if result < 0
            quit
        s.str(string(u#EOL,"Writing Cache ... "))
        repeat n from 0 to result-1
            c.writeByte(n+len, byte[@buffer+n])
            {
            val := c.readByte(n+len)
            if val <> byte[@buffer+n]
               s.str(string(u#EOL,"Write+Readback Error"))
               repeat
            '}
            {
            val := c.readByte(n+len)
            ifnot n // 16
               s.str(string(u#EOL," "))
               s.hex(n+len,4)
            s.out(" ")
            s.hex(val,2)
            '}
        len += result

    f.pclose

    result := \f.popen(filename,"r")
    if result
        s.str(string(u#EOL,"SDCARD File Open for Verify Error ... "))

    len    := 0
    result := 0
    repeat
        result := f.pread(@buffer, BUFSIZE)
        if result < 0
            quit
        repeat n from 0 to result-1
            val := c.readByte(n+len)
            {
            ifnot n // 16
               s.str(string(u#EOL," "))
               s.hex(n+len,4)
            s.out(" ")
            s.hex(val,2)
            '}
            if val <> byte[@buffer+n]
                s.str(string(u#EOL,"Memory Verify Error ... "))
                s.str(string(u#EOL,"Address  "))
                s.hex(n+len,8)
                s.str(string(u#EOL,"Expected "))
                s.hex(byte[@buffer+n],2)
                s.str(string(u#EOL,"Received "))
                s.hex(val,2)
                repeat
        len += result

    s.str(string(u#EOL,"Verified "))
    s.dec(len)
    s.str(string(" of "))
    msize := c.memsize
    s.dec(msize)
    s.str(string(" bytes.",u#EOL))

    xmpbase := c.readWord(6)
    xmvbase := c.readWord(8)
    xmdbase := c.readWord(10)
    xmpcurr := c.readWord(12)
    xmdcurr := c.readWord(14)

    ' clear out VAR memory
    repeat n from xmvbase to xmdcurr-1 step 4
        c.writeLong(n,0)

#endif

    s.str(string($d,$a,"Startup addresses: "))

    s.hex(xmpbase,4)
    s.out(" ")
    s.hex(xmvbase,4)
    s.out(" ")
    s.hex(xmdbase,4)
    s.out(" ")
    s.hex(xmpcurr,4)
    s.out(" ")
    s.hex(xmdcurr,4)
    s.str(string($d,$a))

#ifdef BMA
   d.taskstart(@interpreter,@cachecmd,string("BigSpin"))     'coginit(0, $F004, 0) 'restart cog 0 with the interpreter
   d.start
#else
   coginit(0, @interpreter,@cachecmd)  'restart cog 0 with the interpreter
#endif

CON
        WrByteInst      = $00
        RdByteInst      = $01
        WrWordInst      = $08
        RdWordInst      = $09
        WrLongInst      = $10
        RdLongInst      = $11

        OBJSTART        = $10           ' 32 bit spin would be $1A


DAT                     org
interpreter
long 0 [8]
' BIGSPIN or (notdefined BIGSPIN and notdefined NORMALSPIN)
mboxcmd                 mov     $, par
y                       mov     y, par
op                      add     y, #4
mboxdat                 mov     $, y
x                       mov     x, #$1F0-pbase          'entry, set self to parm count for load
a                       add     y, #4                   'loop
t1                      rdlong  pbase,y
t2                      add     t1, x200                'inc t1 d lsb
adr                     djnz    x,#a
op2                     cogid   id                      'set id
                        mov     dcall,#0

'
'
' Main loop
'
loop                    mov     x,#0                    'reset x

#ifndef BIGSPIN
                        rdbyte  op,pcurr                'get opcode
#else
                        movs    read_byte,#pcurr        ' pcurr is address
                        movd    read_byte_val,#op       ' set op as read destination
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1

                        cmp     op,#$40 wc              'upper?
        if_nc           jmp     #upper

                        mov     t1,op                   ' get load bytecode handler index
                        shr     t1,#2                   ' each handler does up to 4 instructions
                        add     t1,#jumps               ' add jump table offset
                        movs    ladr,t1                 ' set overlay load source address
'
' handle SPIN bytecode calls
'
                        movd    load,   #inst           ' set overlay load destinations
                        movd    load2,  #inst2
                        movd    load3,  #inst3
                        movd    load4,  #inst4
ladr                    mov     t2,0-0
                        call    #load
                        movs    getret, #inst
getflags                test    op,#%01 wz
getc                    test    op,#%10 wc
getret                  ret                             ' getret is modified as jump to bytecode handler


'_____________________________________________________________________________
' Overlay load loop
' NOPs in overlayed code must be non-zero else load will be terminated early
'   Fair compromise for short/long instruction interpretation.
'   Unroll loop 4x for better performance.
'
load
                        rdlong  inst,   t2 wz
        if_nz           add     load,   x800
        if_nz           add     t2,     #4
load2
        if_nz           rdlong  inst2,  t2 wz
        if_nz           add     load2,  x800
        if_nz           add     t2,     #4
load3
        if_nz           rdlong  inst3,  t2 wz
        if_nz           add     load3,  x800
        if_nz           add     t2,     #4
load4
        if_nz           rdlong  inst4,  t2 wz
        if_nz           add     load4,  x800
        if_nz           add     t2,     #4
        if_nz           jmp     #load
load_ret                ret

{
'
' slower version in case we need more space
'
load
                        rdlong  inst,   t2 wz
        if_nz           add     load,   x400
        if_nz           add     t2,     #4
load2
        if_nz           rdlong  inst2,  t2 wz
        if_nz           add     load2,  x400
        if_nz           add     t2,     #4
        if_nz           jmp     #load
load_ret                ret
}

jumps
long    @j0+OBJSTART
long    @j1+OBJSTART
long    @j2+OBJSTART
long    @j3+OBJSTART
long    @j4+OBJSTART
long    @j5+OBJSTART
long    @j6+OBJSTART
long    @j7+OBJSTART
long    @j8+OBJSTART
long    @j9+OBJSTART
long    @jA+OBJSTART
long    @jB+OBJSTART
long    @jC+OBJSTART
long    @jD+OBJSTART
long    @jE+OBJSTART
long    @jF+OBJSTART

inst    long    0
inst2   long    0
inst3   long    0
inst4   long    0 [45-3]          ' longest overlay is 45 instructions

#ifdef BIGSPIN

#define SMALLER

#ifdef SMALLER
'------------------------------------------------------------------------------
' input:  read_mem src = user address to read
' output: readm dest  = data read
' readm instruction can be changed to rdbyte, rdword, rdlong
read_mem                mov     addr, 0-0
                        call    #cache_read
readm                   rdlong  0-0,memp                ' instruction gets modified
read_mem_ret            ret

'------------------------------------------------------------------------------
' input:  write_long src = user address to write
' input:  write_val dest = data to write
' writem instruction can be changed to wrbyte, wrword, wrlong
write_mem               mov     addr, 0-0
                        call    #cache_write
writem                  wrlong  0-0,memp                ' instruction gets modified
write_mem_ret           ret

'------------------------------------------------------------------------------
' input:  read_byte src = user address to read
' output: read_byte_val dest = data read
read_byte               mov     addr, 0-0
                        call    #cache_read
read_byte_val           rdbyte  0-0,memp
read_byte_ret           ret

'------------------------------------------------------------------------------
' input:  read_long src = user address to read
' output: read_long_val dest = data read
read_long               mov     addr, 0-0
                        call    #cache_read
read_long_val           rdlong  0-0,memp
read_long_ret           ret

'------------------------------------------------------------------------------
' input:  write_byte src = user address to write
' input:  write_val dest = data to write
write_byte              mov     addr, 0-0
                        call    #cache_write
write_byte_val          wrbyte  0-0,memp
write_byte_ret          ret

'------------------------------------------------------------------------------
' input:  write_long src = user address to write
' input:  write_val dest = data to write
write_long              mov     addr, 0-0
                        call    #cache_write
write_long_val          wrlong  0-0,memp
write_long_ret          ret

'------------------------------------------------------------------------------
'
' This is cache code for external memory. Cache is write-back.
' Uses a hub cache for get/set data via a single cache line pointer to reduce access time.
' Other methods require 3 hub access per transaction.
' This method requires only one hub access by the user until a cache line miss.
'

cache_write
'                        and     addr, maskword          ' 16 bit LITTLE BIGSPIN only
                        mov     memp, addr              'save address for index
                        andn    addr, #c#CMD_MASK       'ensure a write is not a read
                        or      addr, #c#WRITE_CMD
                        jmp     #cache_access

cache_read
                        and     addr, maskword          ' 16 bit LITTLE BIGSPIN only
                        mov     memp, addr              'save address for index
                        or      addr, #c#READ_CMD       'read must be 3 to avoid needing andn addr,#cache#CMD_MASK

cache_access                                            ' if cacheaddr <> addr, load new cache line
                        wrlong  addr, mboxcmd
                        and     memp, #(c#LINELEN-1)    'memp is index into buffer
:waitres                rdlong  temp, mboxcmd           'wait for cache driver to deliver
                        tjnz    temp,#:waitres          'ok to recycle z flag
                        rdlong  mboxptr, mboxdat        'Get new buffer
                        add     memp, mboxptr           'memp is now HUB buf address of data to read
cache_read_ret
cache_write_ret
cache_access_ret        ret

' cache variables
'
addr                    long    0
temp                    long    0
memp                    long    0
mboxptr                 long    0
'
'------------------------------------------------------------------------------

#else
' not SMALLER

'------------------------------------------------------------------------------
' input:  read_mem src = user address to read
' output: readm dest  = data read
' readm instruction can be changed to rdbyte, rdword, rdlong
read_mem                mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_read
readm                   rdlong  0-0,memp                ' instruction gets modified
                        test    zflag,#1 wz             ' restore zero state
read_mem_ret            ret

'------------------------------------------------------------------------------
' input:  write_long src = user address to write
' input:  write_val dest = data to write
' writem instruction can be changed to wrbyte, wrword, wrlong
write_mem               mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_write
writem                  wrlong  0-0,memp                ' instruction gets modified
                        test    zflag,#1 wz             ' restore zero state
write_mem_ret           ret

'------------------------------------------------------------------------------
' input:  read_byte src = user address to read
' output: read_byte_val dest = data read
read_byte               mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_read
read_byte_val           rdbyte  0-0,memp
                        test    zflag,#1 wz             ' restore zero state
read_byte_ret           ret

'------------------------------------------------------------------------------
' input:  read_long src = user address to read
' output: read_long_val dest = data read
read_long               mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_read
read_long_val           rdlong  0-0,memp
                        test    zflag,#1 wz             ' restore zero state
read_long_ret           ret

'------------------------------------------------------------------------------
' input:  write_byte src = user address to write
' input:  write_val dest = data to write
write_byte              mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_write
write_byte_val          wrbyte  0-0,memp
                        test    zflag,#1 wz             ' restore zero state
write_byte_ret          ret

'------------------------------------------------------------------------------
' input:  write_long src = user address to write
' input:  write_val dest = data to write
write_long              mov     addr, 0-0
                        call    #iscache
        if_ne           call    #cache_write
write_long_val          wrlong  0-0,memp
                        test    zflag,#1 wz             ' restore zero state
write_long_ret          ret

'------------------------------------------------------------------------------
'
' This is cache code for external memory. Cache is write-back.
' Uses a hub cache for get/set data via a single cache line pointer to reduce access time.
' Other methods require 3 hub access per transaction.
' This method requires only one hub access by the user until a cache line miss.
'
iscache
                        muxnz   zflag,#1                'save not zero state - restore not zero at caller
                        and     addr, maskword          ' 16 bit LITTLE BIGSPIN only
                        mov     memp, addr              'ptr + mboxdat = hub address of byte to load
                        and     memp, #(c#LINELEN-1)
                        mov     temp, addr              'ptr + mboxdat = hub address of byte to load
                        andn    temp, #(c#LINELEN-1)
                        cmp     cacheaddr,temp wz       'if cacheaddr == addr, just pull form cache
                        add     memp, mboxptr           'add ptr to memp to get data address
iscache_ret             ret

cache_write
                        mov     memp, addr              'save address for index
                        andn    addr, #c#CMD_MASK       'ensure a write is not a read
                        or      addr, #c#WRITE_CMD
                        jmp     #cache_access

cache_read
                        mov     memp, addr              'save address for index
                        or      addr, #c#READ_CMD       'read must be 3 to avoid needing andn addr,#cache#CMD_MASK

cache_access                                            ' if cacheaddr <> addr, load new cache line
                        wrlong  addr, mboxcmd
                        mov     cacheaddr,addr          'Save new cache address. it's free time here
                        andn    cacheaddr,#c#LINELEN-1  'Kill command bits in free time
:waitres                rdlong  temp, mboxcmd wz        'wait for cache driver to deliver
        if_nz           jmp     #:waitres               'ok to recycle z flag
                        and     memp, #(c#LINELEN-1)    'memp is index into buffer
                        rdlong  mboxptr, mboxdat        'Get new buffer
                        add     memp, mboxptr           'memp is now HUB buf address of data to read
cache_read_ret
cache_write_ret
cache_access_ret        ret

' cache variables
'
addr                    long    0
temp                    long    0
memp                    long    0
mboxptr                 long    0
cacheaddr               long    -1                      ' must start non-zero
'
'------------------------------------------------------------------------------
#endif
' not SMALLER
#endif
' BIGSPIN

'------------------------------------------------------------------------------
' jmpadr routine used by overlaid code
'
jmpadr                  jmpret  getret,#getadrs         'get sign-extended address
        if_nc_and_nz    sub     x,#1                    'if djnz, decrement
                        cmp     x,#1                wc  'determine branch
                        jmp     #apcurr

'------------------------------------------------------------------------------
' regindex routine used by overlaid code
'
regindex
                        or      x,#$1E0                 'install register
                        movs    writef,x
                        movd    writer,x
                        movs    readr,x

        if_c_and_nz     jmp     #mrop                   'register?

        if_nc           call    #popx                   'register bit?
        if_nc           mov     y,x

        if_c            call    #popyx                  'register range?

                        and     x,#$1F                  'trim bit/range
                        and     y,#$1F

                        mov     adr,x                   'get -bitcount into adr
                        sub     adr,y wc                'c=1 if reverse
                        absneg  adr,adr
                        sub     adr,#1

        if_nc           mov     lsb,y                   'get lowest bit into lsb
        if_c            mov     lsb,x

                        muxnc   writev,maskwr           'clear/set reverse
                        muxnc   readv,maskwr

                        or      op,#%1100               'set bit mode
                        jmp     #mrop

'------------------------------------------------------------------------------
' spcurr routine used by overlaid code
'
spcurr  if_nc           mov     pcurr,pbase             'set call pcurr (c=0)
        if_nc           add     pcurr,x
                        jmp     #loop

'------------------------------------------------------------------------------
' apcurr routine used by overlaid code
'
apcurr  if_c_eq_z       add     pcurr,op2               'branch?
                        muxc    op,#%00001001 wz        'if tjz/djnz and not 0, push, else loop
        if_z            jmp     #push                   'if z, push result
        if_nz           jmp     #loop                   'if nz, loop

'------------------------------------------------------------------------------
' pushz routine used by overlaid code
'
pushz
        if_z            jmp     #push                   'if z, push result
        if_nz           jmp     #loop                   'if nz, loop


notx    if_c            xor     x,masklong
                        jmp     #push


'------------------------------------------------------------------------------
'------------------------------------------------------------------------------

'
'
' Upper codes
'
upper                   cmp     op,#$80 wc              'varop?
        if_c            jmp     #varop                  'c=1

                        cmp     op,#$E0 wc              'memop?
        if_c            jmp     #memop

                        mov     a,op    wz              'mathop follows (z=0)
'
'
' Math operation
'                               unary/
'        code        normal     assign  binary   description
'        ---------------------------------------------------------------
'        00000        ->        ->=     b        rotate right
'        00001        <-        <-=     b        rotate left
'        00010        >>        >>=     b        shift right
'        00011        <<        <<=     b        shift left
'        00100        |>        |>=     b        limit minimum (signed)
'        00101        <|        <|=     b        limit maximum (signed)
'        00110        -         -       u        negate
'        00111        !         !       u        bitwise not
'        01000        &         &=      b        bitwise and
'        01001        ||        ||      u        absolute
'        01010        |         |=      b        bitwise or
'        01011        ^         ^=      b        bitwise xor
'        01100        +         +=      b        add
'        01101        -         -=      b        subtract
'        01110        ~>        ~>=     b        shift arithmetic right
'        01111        ><        ><=     b        reverse bits
'        10000        AND               b        boolean and
'        10001        >|        >|      u        encode (0-32)
'        10010        OR                b        boolean or
'        10011        |<        |<      u        decode
'        10100        *         *=      b        multiply, return lower half (signed)
'        10101        **        **=     b        multiply, return upper half (signed)
'        10110        /         /=      b        divide, return quotient (signed)
'        10111        //        //=     b        divide, return remainder (signed)
'        11000        ^^        ^^      u        square root
'        11001        <                 b        test below (signed)
'        11010        >                 b        test above (signed)
'        11011        <>                b        test not equal
'        11100        ==                b        test equal
'        11101        =<                b        test below or equal (signed)
'        11110        =>                b        test above or equal (signed)
'        11111        NOT       NOT     u        boolean not
'
' z = swap binary arguments
'
mathop                  and     a,#%11111               'limit for mbol and mcod

                        ror     mathops,a               'unary or binary?
                        rol     mathops,a        wc

        if_nc           call    #popx                   'pop unary argument
        if_c            call    #popyx                  'pop binary arguments

        if_nc_or_z      xor     x,y                     'if unary or swap, swap x and y
        if_nc_or_z      xor     y,x
        if_nc_or_z      xor     x,y

                        mov     t1,#0

                        test    a,#%10000        wz     'jmp to operation
        if_nc           jmp     #muny
        if_z            jmp     #mcod
                        test    a,#%01000        wc
        if_c            jmp     #mtst

                        test    a,#%00100        wc     'boolean and/or?
        if_nc           cmp     x,#0             wz
        if_nc           muxnz   x,masklong
        if_nc           cmp     y,#0             wz
        if_nc           muxnz   y,masklong
        if_nc           jmp     #mcod2

                        mov     t2,#32                  'multiply/divide
                        abs     x,x              wc
                        muxc    a,#%01100
                        abs     y,y              wc,wz
        if_c            xor     a,#%00100
                        test    a,#%00010        wc
        if_c_and_nz     jmp     #mdiv                   'if divide and y=0, do multiply so result=0

                        shr     x,#1             wc     'multiply
mmul        if_c        add     t1,y             wc
                        rcr     t1,#1            wc
                        rcr     x,#1             wc
                        djnz    t2,#mmul
                        test    a,#%00100        wz
        if_nz           neg     t1,t1
        if_nz           neg     x,x              wz
        if_nz           sub     t1,#1
                        test    a,#%00001        wz
        if_nz           mov     x,t1
                        jmp     #push

mdiv                    shr     y,#1             wc,wz  'divide
                        rcr     t1,#1
        if_nz           djnz    t2,#mdiv
mdiv2                   cmpsub  x,t1             wc
                        rcl     y,#1
                        shr     t1,#1
                        djnz    t2,#mdiv2
                        test    a,#%01000        wc
                        negc    x,x
                        test    a,#%00100        wc
                        test    a,#%00001        wz
        if_z            negc    x,y
                        jmp     #push

mtst                    cmps    x,y              wc,wz  'tests
        if_z            mov     x,#%100                 'equal?
        if_nz           mov     x,#%010                 'above?
        if_c            mov     x,#%001                 'below?
                        andn    x,a              wz
mtst2                   muxz    x,masklong
                        jmp     #push

mcod                    cmp     a,#%01111        wz     'instruction-equivalents
        if_z            neg     y,y
                        mov     t1,a
                        and     t1,#%001100
                        add     a,t1
                        cmp     a,#%011010       wc
        if_nc           sub     a,#%010100
mcod2                   shl     a,#3
                        add     a,#%001000_001
                        movi    mcod3,a
maskpar                 long    $0000FFFC               'nop/constant
mcod3                   ror     x,y                     '(modifying)
                        jmp     #push

muny                    test    a,#%01000        wc     'unaries
        if_nz           jmp     #muny2

                        test    a,#%00001        wz
        if_nc           neg     x,y                     'neg
        if_nc_and_nz    sub     x,#1                    'bitwise not
        if_c            abs     x,y                     'abs
                        jmp     #push

muny2                   test    a,#%00010        wz
        if_c            jmp     #muny3

        if_z            mov     x,#32                   'encode
mncd        if_z        shl     y,#1             wc
        if_z_and_nc     djnz    x,#mncd

        if_nz           mov     x,#1                    'decode
        if_nz           shl     x,y
                        jmp     #push

muny3                   mov     x,#0                    'square root
        if_z            mov     t2,#16
msqr        if_z        shl     y,#1             wc
        if_z            rcl     t1,#1
        if_z            shl     y,#1             wc
        if_z            rcl     t1,#1
        if_z            shl     x,#2
        if_z            or      x,#1
        if_z            cmpsub  t1,x             wc
        if_z            shr     x,#2
        if_z            rcl     x,#1
        if_z            djnz    t2,#msqr

        if_nz           cmp     x,y              wc     'boolean not
        if_nz           muxnc   x,masklong

push
#ifndef BIGSPIN
                        wrlong  x,dcurr                 'push result
#else
                        movs    write_long,#dcurr
                        movd    write_long_val,#x
                        call    #write_long
#endif
                        add     dcurr,#4
                        test    op2,#%01000000   wc     'mathop?
pushret                 jmp     #loop


mathops                 long    %1_01111110_11110101_11111101_0011111
'
'
' Variable operation (c=1)
'
varop                   mov     y,#%10                  'set long

                        mov     adr,op                  'isolate offset
                        and     adr,#%011100

                        test    op,#%100000      wz     'get vbase/dbase

                        jmp     #memopb                 'add base (c=1)
'
'
' Memory operation
'
memop                   mov     y,op                    'set size
                        shr     y,#5
                        and     y,#%11

                        test    op,#%0010000     wc     'index?

        if_c            call    #popx                   'yes, pop and scale
        if_c            shl     x,y

                        test    op,#%0001000     wc     'get base mode
                        test    op,#%0000100     wz

        if_nc_and_z     sub     dcurr,#4                'if no base, pop address
#ifndef BIGSPIN
        if_nc_and_z     rdlong  adr,dcurr
#else
        if_nc_and_z     movs    read_long,#dcurr
        if_nc_and_z     movd    read_long_val,#adr
        if_nc_and_z     call    #read_long
#endif

        if_c_or_nz      movi    sarshr,#%001010_001     'if base, get zero-extended address
        if_c_or_nz      jmpret  getret,#getadrz
                        test    op,#%0001000    wc      'restore c
        if_c_or_nz      mov     adr,op2

                        add     adr,x                  'add any index

        if_nc_and_nz    add     adr,pbase              'if pbase, add
memopb  if_c_and_z      add     adr,vbase              'if vbase, add
        if_c_and_nz     add     adr,dbase              'if dbase, add

                        shl     y,#3                   'set read/write by size
                        movi    writem,y
#ifdef BIGSPIN
                        movi    writemh,y              ' in case it's a HUB write
#endif
                        or      y,#%000000_001
                        movi    readm,y
#ifdef BIGSPIN
                        movi    readmh,y               ' in case it's a HUB read
#endif

                        and     op,#%0011              'set type to memory (followed by mrop)
'
'
' Memory/register operation
'
mrop                    jmpret  getret,#getflags       'get op flags

        if_nc_and_z     jmp     #read                  'read?

        if_nc_and_nz    call    #popx                  'write?
        if_nc_and_nz    jmp     #write

        if_c_and_nz     mov     x,adr                  'address?
'       if_c_and_nz     and     x,maskword
        if_c_and_nz     jmp     #push


                        jmpret  getret,#getop2         'assign, get assignment (c=1)

                        test    op2,#%01111110  wz      'write? (w/push)
        if_z            call    #popx
        if_z            jmp     #:keep

                        jmpret  pushret,#read           'modifier or mathop, read var (c=1 if mathop)

                        test    op2,#%00100000  wz

        if_c            mov     a,op2                   'mathop? set op, z=swap args
        if_c            jmpret  pushret,#mathop         'do math (c=1)
                        sub     dcurr,#4                'unpop var/result (in any case)
        if_c            jmp     #:keep                  'if mathop, write

                        test    op2,#%00010000  wc
        if_nz           jmp     #:incdec

                        test    op2,#%00000100  wz
        if_c            jmp     #:sxcs

                        test    op2,#%00001000  wc
        if_c            jmp     #:rnd
#ifndef BIGSPIN
                        movd    popxr,#t1               'repeat-var loop?
                        call    #popayx                 'pop data (a=to, y=from, t1=step)
                        movd    popxr,#x
#else
                        movs    popxr,#t1               'repeat-var loop?
                        call    #popayx                 'pop data (a=to, y=from, t1=step)
                        movs    popxr,#x
#endif
        if_z            add     dcurr,#4                'if step default, unpop step
        if_z            mov     t1,#1                   'if step default, set step to 1
                        jmpret  getret,#getadrs         'get address
                        cmps    a,y             wc      'reverse range?
                        sumc    x,t1                    'add/sub step to/from var
                        call    #range                  'check if x in range y..a according to c
        if_nc           add     pcurr,op2               'if in range, branch
                        jmp     #:restore

:rnd                    min     x,#1                    '?var/var?
                        mov     y,#32
                        mov     a,#%10111
        if_nz           ror     a,#1
:rndlp                  test    x,a wc
        if_z            rcr     x,#1
        if_nz           rcl     x,#1
                        djnz    y,#:rndlp      wc       'c=0
                        jmp     #:stack

:sxcs                   test    op2,#%00001000 wc
        if_nc_and_z     shl     x,#24                   '~var/~~var
        if_nc_and_z     sar     x,#24
        if_nc_and_nz    shl     x,#16
        if_nc_and_nz    sar     x,#16
        if_c            muxnz   x,masklong              'var~/var~~
                        jmp     #:stack

:incdec                 sumc    x,#1                    '++var/var++/--var/var--
                        test    op2,#%00000100 wc       'mask result by size
                        test    op2,#%00000010 wz
        if_nc_and_z     rev     x,adr
        if_nc_and_z     rev     x,adr
        if_nc_and_nz    and     x,#$FF
        if_c_and_z      and     x,maskword
                        test    op2,#%00001000 wc       'pre-inc/dec or post-inc/dec?

:stack
#ifndef BIGSPIN
        if_nc           wrlong  x,dcurr                 'if not var~/var~~/var++/var--, write stack
#else
        if_nc           movs    write_long,#dcurr
        if_nc           movd    write_long_val,#x
        if_nc           call    #write_long
#endif

:keep                   test    op2,#%10000000 wc       'keep value on stack?
        if_c            add     dcurr,#4                '(followed by write)

:restore                movs    pushret,#loop           'restore pushret, followed by write
'
'
' Write memory/register
'
write                   test    op,#%1100 wc,wz         'get type into flags

#ifndef BIGSPIN
writem  if_z            wrlong  x,adr                   'memory?
        if_z            jmp     #loop
#else
        if_nz           jmp     #writenext

                        muxc    carry,#1                ' save carry state
                        cmp     adr,HUB_ADDR wc         ' if adr > HUB_ADDR set carry
        if_c            movs    write_mem,#adr
        if_c            movd    writem,#x
        if_c            call    #write_mem              ' write to external memory
writemh if_nc           wrlong  x,adr                   ' write to hub
                        test    carry,#1 wc             ' restore carry state
                        jmp     #loop
writenext
#endif
        if_nc           neg     a,#1                    'register field?
        if_nc           rev     a,adr
        if_nc           shl     a,lsb
        if_nc           xor     a,masklong
        if_nc           rev     x,adr
writev  if_nc           rev     x,adr
        if_nc           shl     x,lsb
writef  if_nc           and     a,$1FF
        if_nc           or      x,a
writer                  mov     $1FF,x                  'register - modified
                        jmp     #loop
'
'
' Read memory/register
'
read                    test    op,#%1100 wc,wz         'get type into flags

#ifndef BIGSPIN
readm   if_z            rdlong  x,adr                   'memory?
        if_z            jmp     #push
#else
        if_nz           jmp     #readnext

                        muxc    carry,#1                ' save carry state
                        cmp     adr,HUB_ADDR wc         ' if adr < HUB_ADDR set carry
        if_c            movs    read_mem,#adr
        if_c            movd    readm,#x
        if_c            call    #read_mem               ' read from external memory
readmh  if_nc           rdlong  x,adr                   ' read from HUB
                        test    carry,#1 wc             ' restore carry state
                        jmp     #push

readnext
#endif

readr                   mov     x,$1FF                  'register - modified
        if_nc           shr     x,lsb                   'register field?
        if_nc           rev     x,adr
readv   if_nc           rev     x,adr                   ' modified
                        jmp     #push
'
'
' Get address
'
getadrs                 movi    sarshr,#%001110_001     'set sign-extended

getadrz
#ifndef BIGSPIN
                        rdbyte  op2,pcurr               'get first byte
#else
                        movs    read_byte,#pcurr        ' pcurr is address
                        movd    read_byte_val,#op2      ' set op as read destination
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1
                        test    op2,#$80        wc      'if bit7 set, another byte

                        shl     op2,#25                 'sign/zero-extend from bit6
sarshr                  sar     op2,#25                 'sar for jX / shr for memop

getop2
#ifndef BIGSPIN
        if_c            rdbyte  t2,pcurr                'if another byte, get second byte and shift in
#else
        if_c            movs    read_byte,#pcurr        ' pcurr is address
        if_c            movd    read_byte_val,#t2       ' set op as read destination
        if_c            call    #read_byte              ' get byte
#endif
        if_c            add     pcurr,#1
        if_c            shl     op2,#8
        if_c            or      op2,t2

                        jmp     #getc                   'restore c for jX
'
'
' Pops
'
#ifndef BIGSPIN
popayx                  sub     dcurr,#4
                        rdlong  a,dcurr

popyx                   sub     dcurr,#4
                        rdlong  y,dcurr

popx                    sub     dcurr,#4
popxr                   rdlong  x,dcurr
#else
popayx                  sub     dcurr,#4
                        movs    read_long,#dcurr
                        movd    read_long_val,#a
                        call    #read_long

popyx                   sub     dcurr,#4
                        movs    read_long,#dcurr
                        movd    read_long_val,#y
                        call    #read_long

popx                    sub     dcurr,#4
                        movs    read_long,#dcurr
popxr                   movd    read_long_val,#x        ' modified by repeat-var
                        call    #read_long
#endif

popx_ret
popyx_ret
popayx_ret              ret
'
'
' Check range
' must be preceded by:
'                       cmps    a,y                wc
'
range        if_c       xor     a,y                     'if reverse range, swap range values
        if_c            xor     y,a
        if_c            xor     a,y

                        cmps    x,y                wc   'c=0 if x within range
        if_nc           cmps    a,x                wc

range_ret               ret

x200                    long    $200
x800                    long    $800
maskpar32               long    $FFFFFFFC               'nop/constant

carry                   long    0
zflag                   long    0
HUB_ADDR                long    b#HUB_ADDR
cogptr                  long    @cogbuf+$10
bufp                    long    0
:fit                    long    $0[$1E5 - $]            'fill so pcurr, etc are at right address.
                        fit     $1E5
'
'
' Constants
'
maskword                long    $0000FFFF                'nop/constant
masklong                long    $FFFFFFFF
maskwr                  long    $00800000
'
'
' Variables
'
lsb                     res     1
id                      res     1
dcall                   res     1
pbase                   res     1
vbase                   res     1
dbase                   res     1
pcurr                   res     1
dcurr                   res     1


'==================================================================================
' end cog resident code
' start overlay code
'==================================================================================

'
'
' drop anchor           ' could leave in cog ....
'
long                    0       ' end of method

dat     org inst

j0                      or        op,pbase              'add pbase into flags

#ifndef BIGSPIN
                        wrword  op,dcurr                'push return pbase (and flags)
                        add     dcurr,#2

                        wrword  vbase,dcurr             'push return vbase
                        add     dcurr,#2

                        wrword  dbase,dcurr             'push return dbase
                        add     dcurr,#2

                        wrword  dcall,dcurr             'push dcall (later used for pcurr)
#else
                        movi    writem,#WrLongInst
                        movs    write_mem,#dcurr        'source is always dcurr here

                        'wrword  op,dcurr               'push return pbase (and flags)
                        movd    writem,#op
                        call    #write_mem
                        add     dcurr,#4

                        'wrword  vbase,dcurr            'push return vbase
                        movd    writem,#vbase
                        call    #write_mem
                        add     dcurr,#4

                        'wrword  dbase,dcurr            'push return dbase
                        movd    writem,#dbase
                        call    #write_mem
                        add     dcurr,#4

                        'wrword  dcall,dcurr            'push dcall (later used for pcurr)
                        movd    writem,#dcall
                        call    #write_mem
#endif
                        mov     dcall,dcurr             'set new dcall
                        add     dcurr,#4

                        jmp        #push                'init 'result' to 0

long                    0       ' end of method

'
'
' jmp
' call sub
' call obj.sub
' call obj[].sub
'
dat     org inst

j1                      jmp     #j11
callobj                 add     pbase,x                 'set relative obj bases
                        add     vbase,y         wc      '(c=0, z=0 still)

j11     if_nc_and_z     jmp     #jmpadr                 'entry, jmp?

        if_c_and_nz     call    #popx                   'pop index

#ifndef BIGSPIN
                        rdbyte  y,pcurr                 'get obj/sub byte
#else
                        movs    read_byte,#pcurr        ' pcurr is address
                        movd    read_byte_val,#y
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1

        if_c_and_nz     add     y,x                     'add any obj index

                        shl     y,#2                    'lookup words from table
                        add     y,pbase
#ifndef BIGSPIN
                        rdlong  y,y
                        mov     x,y                     'get low word
                        shr     y,#16                   'get high word
#else
                        movi    readm,#RdWordInst

                        movs    read_mem,#y            ' pcurr is address
                        movd    readm,#x
                        call    #read_mem              ' get byte

                        add     y,#2

                        movs    read_mem,#y            ' pcurr is address
                        movd    readm,#y
                        call    #read_mem              ' get byte
#endif

        if_c            jmp     #callobj        wz      'obj[].sub? (z=0)

                        mov     dbase,dcall             'get new dcall
#ifndef BIGSPIN
                        rdword  dcall,dcall             'set old dcall
                        wrword  pcurr,dbase             'set return pcurr
                        add     dbase,#2                'set call dbase
#else
                        movi    readm,#RdWordInst
                        movs    read_mem,#dcall
                        movd    readm,#dcall
                        call    #read_mem

                        movi    writem,#WrWordInst
                        movs    write_mem,#dbase
                        movd    writem,#pcurr
                        call    #write_mem
                        add     dbase,#4                'set call dbase
#endif

                        add     dcurr,y                 'set call dcurr
                        jmp     #spcurr

long                    0       ' end of method
'
'
' tjz - push if nz
' djnz - push if nz
' jz
' jnz
'
dat     org inst

j2                      call    #popx                   'pop count/boolean
                        jmp     #jmpadr

long                    0       ' end of method
'
'
' casedone
' value case
' range case
' lookdone
'
dat     org inst

j3      if_nc_and_z     jmp     #cased                  'if casedone, pop target+address, jump

        if_c_and_nz     sub     dcurr,#12               'if lookdone, pop target+address+index
        if_c_and_nz     jmp     #push                   '..and push zero

                        jmpret  getret,#getadrs wz      'case, get sign-extended address (c=same, z=0)
'
'
' value lookup
' value lookdown
' range lookup
' range lookdown
'
' copy of j4 needed by j3
'
_j4     if_nc           call    #popyx                  'if value, pop value and target
        if_nc           mov     a,y                     'x = target, y..a = range

        if_c            call    #popayx                 'if range, pop range values and target

                        sub     dcurr,#8                'pop index into t1 (underneath address)
#ifndef BIGSPIN
                        rdlong  t1,dcurr
#else
                        movs    read_long,#dcurr
                        movd    read_long_val,#t1
                        call    #read_long
#endif
                        cmps    a,y      wc             'reverse range?

                        negc    t2,x                    'set t2 for in-range..
        if_nc           sumz    t2,t1                   'forward lookup:        -t1 + x + y
        if_c            add     t2,t1                   'forward lookdown:         t1 + x - y
        if_nc           sumnz   t2,y                    'reverse lookup:         t1 - x + y
        if_c            add     t2,y                    'reverse lookdown:          t1 - x + y

                        call    #range                  'check if x in range y..a according to c

        if_z            cmp     x,t1     wc             'if lookup, c=0 if index within range
                        add     t1,a                    'on first compare, t1 = index
                        sub     t1,y                    'on second compare, t1 = index+a-y
        if_z_and_nc     cmp     t1,x     wc

        if_nc           mov     t1,t2                   'if in range, t1 = t2
        if_c            add     t1,#1                   'if out of range, t1 = index+a-y+1

                        test    op,#%100 wz             'look or case?
#ifndef BIGSPIN
        if_z            wrlong  t1,dcurr                'if look, update index
#else
        if_z            movs    write_long,#dcurr       'if look, update index
        if_z            movd    write_long,#t1          'if look, update index
        if_z            call    #write_long
#endif

                        add     dcurr,#12               'unpop index+address+target

cased   if_z_and_nc     call    #popyx                  'if look true or casedone, pop target and address

         if_z           jmp     #spcurr                 'if look true or casedone, branch or loop
                        jmp     #apcurr                 'case, branch if true

long                    0       ' end of method
'
'
' value lookup
' value lookdown
' range lookup
' range lookdown
'
'
dat     org inst

j4      if_nc           call    #popyx                  'if value, pop value and target
        if_nc           mov     a,y                     'x = target, y..a = range

        if_c            call    #popayx                 'if range, pop range values and target

                        sub     dcurr,#8                'pop index into t1 (underneath address)
#ifndef BIGSPIN
                        rdlong  t1,dcurr
#else
                        movs    read_long,#dcurr
                        movd    read_long_val,#t1
                        call    #read_long
#endif

                        cmps    a,y             wc      'reverse range?

                        negc    t2,x                    'set t2 for in-range..
        if_nc           sumz    t2,t1                   'forward lookup:        -t1 + x + y
        if_c            add     t2,t1                   'forward lookdown:         t1 + x - y
        if_nc           sumnz   t2,y                    'reverse lookup:         t1 - x + y
        if_c            add     t2,y                    'reverse lookdown:          t1 - x + y

                        call    #range                  'check if x in range y..a according to c

        if_z            cmp     x,t1    wc              'if lookup, c=0 if index within range
                        add     t1,a                    'on first compare, t1 = index
                        sub     t1,y                    'on second compare, t1 = index+a-y
        if_z_and_nc     cmp     t1,x    wc

        if_nc           mov     t1,t2                   'if in range, t1 = t2
        if_c            add     t1,#1                   'if out of range, t1 = index+a-y+1

                        test    op,#%100 wz             'look or case?

#ifndef BIGSPIN
        if_z            wrlong  t1,dcurr                'if look, update index
#else
        if_z            movs    write_long,#dcurr       'if look, update index
        if_z            movd    write_long,#t1          'if look, update index
        if_z            call    #write_long
#endif

                        add     dcurr,#12               'unpop index+address+target

        if_z_and_nc     call    #popyx                  'if look true or casedone, pop target and address

        if_z            jmp     #spcurr                 'if look true or casedone, branch or loop
                        jmp     #apcurr                 'case, branch if true


long                    0       ' end of method
'
'
' pop
' run
' STRSIZE(string)
' STRCOMP(stringa,stringb)
'
dat     org inst

j5      if_z            call    #popx                   'if pop/strsize, pop count/string
        if_nc_and_z     sub     dcurr,x                 'if pop, subtract count from dcurr

        if_nc_and_nz    mov     lsb,pcurr               'if run, save pcurr and set to $FFFC
        if_nc_and_nz    mov     pcurr,maskpar

        if_nc           jmp     #loop                   'if pop/run, loop


        if_nz           call    #popyx                  'if strcomp, pop stringa/stringb
                        mov     a,x                     'stringa/string in a, stringb in y
                        neg     x,#1                    'init !count in x

j5loop
#ifndef BIGSPIN
                        rdbyte  t1,a                    'measure/compare string(s)
#else
                        movs    read_byte,#a
                        movd    read_byte_val,#t1
                        call    #read_byte              ' get byte
#endif
                        add     a,#1
        if_z            tjz     t1,#notx                'if strsize and 0, return count (c=1)
#ifndef BIGSPIN
        if_nz           rdbyte  t2,y
#else
        if_nz           movs    read_byte,#y
        if_nz           movd    read_byte_val,#t2
        if_nz           call    #read_byte              ' get byte
#endif
        if_nz           add     y,#1
        if_nz           xor     t1,t2
        if_nz           tjnz    t1,#mtst2               'if strcomp and mismatch, return false (z=0)
        if_nz           tjz     t2,#mtst2        wz     'if strcomp and 0, return true (z=1)
                        djnz    x,#j5loop               'loop - never falls through

long                    0       ' end of method

'
'
' BYTEFILL(start,value,count)
' WORDFILL(start,value,count)
' LONGFILL(start,value,count)
'
' BYTEMOVE(to,from,count)
' WORDMOVE(to,from,count)
' LONGMOVE(to,from,count)
'
'
dat     org inst

j6
j7                      call    #popayx                 'fill/move/wait, pop parameters
        if_c_and_nz     jmp     #waitpin                'waitpeq/waitpne?

                        tjz     a,#loop                 'if count=0, done

                        test    op,#%100        wc,wz   'fill or move?
                        and     op,#%11                 'isolate size bits
        if_z            mov     t1,y                    'if fill, set value

        if_nz           cmp     y,x                wc   'if upward move, modify pointers
        if_nz_and_c     mov     t2,a
        if_nz_and_c     sub     t2,#1
        if_nz_and_c     shl     t2,op
        if_nz_and_c     add     y,t2
        if_nz_and_c     add     x,t2

                        negc    t2,#1                   'set inc/dec
                        shl     t2,op

                        shl     op,#3                   'set word size
#ifndef BIGSPIN
                        movi    fill,op
                        or      op,#%000000_001
                        movi    move,op
                        long    $f00                    ' can't use nop
#else
                        movi    writem,op
                        movi    fillhub,op
                        or      op,#%000000_001
                        movi    readm,op
                        movi    movehub,op
#endif

#ifndef BIGSPIN
move    if_nz           rdlong  t1,y                    'move/fill loop
        if_nz           add     y,t2
fill                    wrlong  t1,x
#else
                        muxc    carry,#1
move    if_nz           cmp     HUB_ADDR,y wc           ' if D < S movehub
        if_nz_and_nc    movs    read_mem,#y
        if_nz_and_nc    movd    readm,#t1
        if_nz_and_nc    call    #read_mem
movehub if_nz_and_c     rdlong  t1,y
        if_nz           add     y,t2
                        cmp     HUB_ADDR,x wc           ' if D < S fillhub
        if_nc           movs    write_mem,#x
        if_nc           movd    writem,#t1
        if_nc           call    #write_mem
fillhub if_c            wrlong  t1,x
#endif
                        add     x,t2
                        djnz    a,#move
                        test    carry,#1 wc             ' restore carry
                        jmp     #loop
'
'
' WAITPEQ(data,mask,port)
' WAITPNE(data,mask,port)
'
waitpin                 test    a,#1                wc
                        test    op,#%100        wz

        if_z            waitpeq x,y                     'waitpeq
        if_nz           waitpne x,y                     'waitpne

                        jmp     #loop

long                    0       ' end of method
'
'
' CLKSET(mode,freq)
' COGSTOP(id)
' LOCKRET(id)
' WAITCNT(count)
'
dat     org inst

j8      if_nc_and_z     call    #popyx                  'clkset
        if_nc_and_z     clkset  x
        if_nc_and_z     wrlong  y,#$0000                ' these always access hub
        if_nc_and_z     wrbyte  x,#$0004

        if_c_or_nz      call    #popx                   'pop parameter
#ifdef BMA
        if_nc_and_nz    add     x,#1
#endif
        if_nc_and_nz    cogstop x                       'cogstop
        if_c_and_z      lockret x                       'lockret
        if_c_and_nz     waitcnt x,#0                    'waitcnt

                        jmp     #loop

long                    0       ' end of method
'
'
' SPR[nibble]
' WAITVID(colors,pixels)
'
dat     org inst

j9
        if_c_and_nz     call    #popyx                  'waitvid
        if_c_and_nz     waitvid x,y
        if_c_and_nz     jmp     #loop

                        call    #popx                   'spr
                        or      x,#$10
                        test    x,#$10  wc,wz           'c=1, z=0

                        jmp     #regindex

long                    0       ' end of method
'
'
' COGINIT(id,ptr,par)
' LOCKNEW
' LOCKSET(id)
' LOCKCLR(id)
'
dat     org inst

jA
jB      if_c            jmp     #:lock                  'lockclr/lockset?

        if_z            call    #popayx                 'coginit, pop parameters
        if_z            and     a,maskpar               'assemble fields
        if_z            shl     a,#16
        if_z            and     y,maskpar
#ifdef BIGSPIN
        if_z            call    #loadPASM
#endif
        if_z            shl     y,#2
        if_z            or      y,a
        if_z            max     x,#8
        if_z            or      x,y
        if_z            coginit x                wc,wr

        if_nz           locknew x                wc     'locknew

        if_c            neg     x,#1                    '-1 if c, else 0..7
                        jmp     #:push


:lock                   call    #popx                   'lockclr/lockset, pop id

        if_z            lockset x                wc     'clr/set lock
        if_nz           lockclr x                wc

                        muxc    x,masklong              '-1 if c, else 0


:push                   test    op,#%100        wz      'push result?
                        jmp     #pushz


'
' LoadPASM code from external memory for launching COG
'
loadPASM
                        mov     adr,y
                        mov     bufp,cogptr
                        mov     y,cogptr
                        mov     t1,#$1f0
:loop
                        movs    read_long,#adr
                        movd    read_long_val,#t2
                        call    #read_long
                        wrlong  t2,bufp
                        add     adr,#4
                        add     bufp,#4
                        djnz    t1,#:loop
loadPASM_ret            ret


long                    0       ' end of method
'
'
' ABORT
' ABORT value
' RETURN
' RETURN value
'
dat     org inst

jC
#ifndef BIGSPIN
        if_z            rdlong  x,dbase                 'if no value, get 'result'
#else
        if_z            movs    read_long,#dbase
        if_z            movd    read_long_val,#x
        if_z            call    #read_long
#endif
        if_nz           call    #popx                   'if value, pop result

:loop                   mov     dcurr,dbase             'restore dcurr

#ifndef BIGSPIN
                        sub     dcurr,#2                'pop pcurr
                        rdword  pcurr,dcurr             ' pcurr ok ?

                        sub     dcurr,#2                'pop dbase
                        rdword  dbase,dcurr

                        sub     dcurr,#2                'pop vbase
                        rdword  vbase,dcurr

                        sub     dcurr,#2                'pop pbase (and flags)
                        rdword  pbase,dcurr
#else
                        movi    readm,#RdLongInst
                        movs    read_mem,#dcurr         ' source is always dcurr here

                        sub     dcurr,#4                'pop pcurr
                        'rdword  pcurr,dcurr            ' pcurr ok ?
                        movd    readm,#pcurr
                        call    #read_mem

                        sub     dcurr,#4                'pop dbase
                        'rdword  dbase,dcurr
                        movd    readm,#dbase
                        call    #read_mem

                        sub     dcurr,#4                'pop vbase
                        'rdword  vbase,dcurr
                        movd    readm,#vbase
                        call    #read_mem

                        sub     dcurr,#4                'pop pbase (and flags)
                        'rdword  pbase,dcurr
                        movd    readm,#pbase
                        call    #read_mem
#endif

        if_nc           test    pbase,#%10        wc    'if abort and !try, return again
        if_nc           jmp     #:loop

                        test    pbase,#%01        wz    'get push flag

                        and     pbase,maskpar32         'trim pbase

                        jmp     #pushz                  'push 'result'?

long                    0       ' end of method
'
'
' constant -1..1
' constant mask
'
dat     org inst

jD                      mov     x,op                    'get -1..2 into x
                        sub     x,#%00110101

        if_nc_or_z      jmp     #push                   'if constant -1..1, push


#ifndef BIGSPIN
                        rdbyte  y,pcurr                 'constant mask, get data byte
#else
                        movs    read_byte,#pcurr
                        movd    read_byte_val,#y
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1

                        rol     x,y                     'decode, x = 2 before

                        test    y,#%001_00000        wc 'decrement?
        if_c            sub     x,#1

                        test    y,#%010_00000        wc 'not?
        if_c            jmp     #notx
                        jmp     #push

long                    0       ' end of method
'
'
' constant 1..4 bytes
'
dat     org inst

jE                      sub     op,#%00111000-1

:loop
#ifndef BIGSPIN
                        rdbyte  a,pcurr
#else
                        movs    read_byte,#pcurr
                        movd    read_byte_val,#a
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1
                        shl     x,#8
                        or      x,a
                        djnz    op,#:loop

                        jmp     #push

long                    0       ' end of method
'
'
' <unused>
' register[bit] op
' register[bit..bit] op
' register op
'
dat     org inst

jF
#ifndef BIGSPIN
                        rdbyte  x,pcurr                 'register, get reg+op byte
#else
                        movs    read_byte,#pcurr
                        movd    read_byte_val,#x
                        call    #read_byte              ' get byte
#endif
                        add     pcurr,#1

                        mov     op,x                    'justify op (sets type to register)
                        shr     op,#5

                        jmp     #regindex

long                    0       ' end of method

jX

